#import "PostTracker.h"
#import "Post.h"


static PostTracker *globalTracker = nil;
static const int kPostLifetime = 3600*24*20;  // 20 days

@implementation PostTracker
- (PostTracker *)init {
	if (self = [super init]) {
	}
	return self;
}
- (BOOL)postInPosts:(NSArray *)array hasPostId:(NSString *)postId {
	for (Post *p in array) {
		if ([postId isEqualToString:p.postId]) {
			return TRUE;
		}
	}
	return FALSE;
}

- (NSArray *)getPostsForPredicate:(NSPredicate *)predicate error:(NSError **)errorRef; {
	NSFetchRequest *request = [[NSFetchRequest alloc] init];
	[request setEntity:[NSEntityDescription entityForName:@"Post" inManagedObjectContext:[self managedObjectContext]]];
	[request setPredicate:predicate];
	NSArray *fetchResults = [[self managedObjectContext] executeFetchRequest:request error:errorRef];
	[request release];
	return fetchResults;
}

- (NSArray *)getPostsForThread:(NSString *)threadId error:(NSError **)errorRef {
	return [self getPostsForPredicate:[NSPredicate predicateWithFormat:@"threadId = %@", threadId] error:errorRef];
}

- (NSArray *)getPostsBeforeTime:(NSDate *)time error:(NSError **)errorRef {
	return [self getPostsForPredicate:[NSPredicate predicateWithFormat:@"touchTime < %@", time] error:errorRef];
}

- (enum PostTrackerPostNovelty)getPostNovelty:(NSString *)postId inThread:(NSString *)threadId {
	NSError *error;
	NSArray *fetchResults = [self getPostsForThread:threadId error:&error];
	if (fetchResults == nil) {
		NSLog(@"Error: %@ (%@)", error, [error userInfo]);
		return PostTrackerPostNoveltyNotNew;
	}
	if ([self postInPosts:fetchResults hasPostId:postId]) {
		return PostTrackerPostNoveltyNotNew;
	}
	return PostTrackerPostNoveltyNew;
}

- (void)noticePost:(NSString *)postId inThread:(NSString *)threadId {
	Post *post = (Post *)[NSEntityDescription insertNewObjectForEntityForName:@"Post" inManagedObjectContext:[self managedObjectContext]];
	post.postId = postId;
	post.threadId = threadId;
	post.touchTime = [NSDate date];
	NSError *error;
	if (![[self managedObjectContext] save:&error]) {
		NSLog(@"Error saving: %@ (%@)", error, [error userInfo]);
	}
}

- (NSInteger)knownCommentsForThread:(NSString *)threadId {
	NSError *error;
	NSArray *fetchResults = [self getPostsForThread:threadId error:&error];
	if (fetchResults == nil) {
		NSLog(@"Error: %@ (%@)", error, [error userInfo]);
		return 0;
	}
	return [fetchResults count];
}

+ (PostTracker *)thePostTracker {
	if (globalTracker == nil) {
		globalTracker = [[PostTracker alloc] init];
	}
	return globalTracker;
}

- (void)dumpChangesAndCullOldies {
	NSError *error = nil;
	if (managedObjectContext != nil) {
		NSArray *posts = [self getPostsBeforeTime:[NSDate dateWithTimeIntervalSinceNow:-kPostLifetime]
		                                    error:&error];
		for (Post *item in posts) {
			[managedObjectContext deleteObject:item];
		}

		if ([managedObjectContext hasChanges] && ![managedObjectContext save:&error]) {
			NSLog(@"Unresolved error %@, %@", error, [error userInfo]);
		}
	}
}

/**
 applicationWillTerminate: saves changes in the application's managed object context before the application terminates.
 */
- (void)applicationWillTerminate:(UIApplication *)application {
	[self dumpChangesAndCullOldies];
}
- (void)applicationDidEnterBackground:(UIApplication *)application {
	[self dumpChangesAndCullOldies];
}

#pragma mark -
#pragma mark Core Data stack

/**
 Returns the managed object context for the application.
 If the context doesn't already exist, it is created and bound to the persistent store coordinator for the application.
 */
- (NSManagedObjectContext *) managedObjectContext {

	if (managedObjectContext != nil) {
		return managedObjectContext;
	}

	NSPersistentStoreCoordinator *coordinator = [self persistentStoreCoordinator];
	if (coordinator != nil) {
		managedObjectContext = [[NSManagedObjectContext alloc] init];
		[managedObjectContext setPersistentStoreCoordinator: coordinator];
	}
	return managedObjectContext;
}


/**
 Returns the managed object model for the application.
 If the model doesn't already exist, it is created by merging all of the models found in the application bundle.
 */
- (NSManagedObjectModel *)managedObjectModel {

	if (managedObjectModel != nil) {
		return managedObjectModel;
	}
	managedObjectModel = [[NSManagedObjectModel mergedModelFromBundles:nil] retain];
	return managedObjectModel;
}


/**
 Returns the persistent store coordinator for the application.
 If the coordinator doesn't already exist, it is created and the application's store added to it.
 */
- (NSPersistentStoreCoordinator *)persistentStoreCoordinator {

	if (persistentStoreCoordinator != nil) {
		return persistentStoreCoordinator;
	}

	NSURL *storeUrl = [NSURL fileURLWithPath: [[self applicationDocumentsDirectory] stringByAppendingPathComponent: @"HNWeb.sqlite"]];

	NSError *error = nil;
	persistentStoreCoordinator = [[NSPersistentStoreCoordinator alloc] initWithManagedObjectModel:[self managedObjectModel]];
	if (![persistentStoreCoordinator addPersistentStoreWithType:NSSQLiteStoreType configuration:nil URL:storeUrl options:nil error:&error]) {
		NSLog(@"Unresolved error %@, %@", error, [error userInfo]);
		abort();
	}

	return persistentStoreCoordinator;
}


#pragma mark -
#pragma mark Application's Documents directory

/**
 Returns the path to the application's Documents directory.
 */
- (NSString *)applicationDocumentsDirectory {
	return [NSSearchPathForDirectoriesInDomains(NSDocumentDirectory, NSUserDomainMask, YES) lastObject];
}


- (void)dealloc {
	[managedObjectContext release];
	[managedObjectModel release];
	[persistentStoreCoordinator release];

	[super dealloc];
}
@end
