#import "consumption.h"

// Many of these functions take an interval [pos, end) that represents a
// subset of a file, and their goal is to advance pos forward.  Sometimes
// they'll advance pos to end, but other times they'll stop slightly
// short of end, when they want to see more bytes on the end of the stream.

// Note that a "segment" is an HTML tag (begun and ended by angle brackets)
// or a text node.



// tryProgress(pos, end, s)
//
// Tries matching the string s with the front of [pos, end).  Returns NULL
// if s does not match, pos if end is reached, pos+strlen(s) if s is
// successfully matched.
const char *tryProgress(const char *pos, const char *end, const char *s) {
	const char *w = pos;
	for (;;) {
		if (!*s) {
			return w;
		}
		if (w == end) {
			return pos;
		}
		if (*w != *s) {
			return NULL;
		}
		++w;
		++s;
	}
}

// skipThroughString(posRef, end, s)
//
// Finds the string s in [*posRef, end) and modifies posRef so that points after the
// string, returning the pointer to the beginning of said string.  Otherwise, returns
// NULL and skips posRef as far forward as possible without potentially skipping
// an instance of s.
const char *skipThroughString(const char **posRef, const char *end, const char *s) {
	if (*s == '\0') {
		return *posRef;
	}
	char firstChar = *s;
	const char *pos = *posRef;
	while (pos < end) {
		if (*pos == firstChar) {
			const char *afterProgress = tryProgress(pos, end, s);
			if (afterProgress) {
				*posRef = afterProgress;
				return (afterProgress != pos) ? pos : NULL;
			} else {
				++pos;
			}
		} else {
			++pos;
		}
	}
	*posRef = pos;
	return NULL;
}

// skipAnyWhitespace(pos, end)
//
// Returns the first address past the isspace characters at the
// front of [pos, end).
const char *skipAnyWhitespace(const char *pos, const char *end) {
	while (pos < end && isspace(*pos)) {
		++pos;
	}
	return pos;
}

// eatToOneOf(pos, end, cs)
//
// Returns the first address past the characters not in cs at the
// front of [pos, end).  Or returns NULL.
const char *eatToOneOf(const char *pos, const char *end, const char *cs) {
	while (pos < end) {
		if (strchr(cs, *pos)) {
			return pos;
		}
		++pos;
	}
	return NULL;
}

// eatToChar(pos, end, c)
//
// Returns the address to the first character equal to c in [pos, end).
// Or returns NULL.
const char *eatToChar(const char *pos, const char *end, char c) {
	return memchr(pos, c, end - pos);
}

// eatThroughChar(pos, end, c)
//
// Returns 1 + eatToChar(pos, end, c), unless it returns NULL, in which
// case this returns NULL.
const char *eatThroughChar(const char *pos, const char *end, char c) {
	const char *p = eatToChar(pos, end, c);
	return p != NULL ? p + 1 : NULL;
}

// eatSegment(pos, end)
//
// Returns the address at the end of the segment at the front of [pos, end), or
// returns NULL.
const char *eatSegment(const char *pos, const char *end) {
	if (pos == end) {
		return NULL;
	}
	if (*pos == '<') {
		return eatThroughChar(pos, end, '>');
	}
	else {
		return eatToChar(pos, end, '<');
	}
}

// skipThroughSegment(posRef, end, s)
//
// Skips through the segment matching the given string.  Returns the address
// of the beginning of the last segment, modifying *posRef to point to the
// end of the segment.  Returns NULL if unsuccessful, advancing posRef as far
// as it can.
const char *skipThroughSegment(const char **posRef, const char *end, const char *s) {
	const char *pos = *posRef;
	for (;;) {
		const char *nextPos = eatSegment(pos, end);
		if (!nextPos) {
			*posRef = pos;
			return NULL;
		}
		if (regionEqualsString(pos, nextPos, s)) {
			*posRef = nextPos;
			return pos;
		}
		pos = nextPos;
	}
}

// segmentTagDirection(beg, end, tagName)
//
// Tells the "direction" of the tag contained in [beg, end).  It is expected
// that beg[0] == '<' and that end[-1] == '>'.
// 
// Returns -1 for matching end tags, 0 for other segments, 1 for matching
// start tags.  A "matching tag" is one with the given tag name.
int segmentTagDirection(const char *beg, const char *end, const char *tagName) {
	size_t segSize = end - beg;
	if (segSize < 3 || *beg != '<') {
		return 0;
	}
	BOOL endTag = (beg[1] == '/');
	const char *scannedTagName = stringPrefixWalk(beg + 1 + endTag, end - 1, tagName);
	if (scannedTagName && !isalpha(*scannedTagName)) {
		return endTag ? -1 : (end[-2] == '/' ? 0 : 1);
	}
	return 0;
}

// skipThroughCloser(posRef, end, tagName)
//
// Skips through the next closing tag in [*posRef, end) that matches the
// start tag with tagName that we have already skipped.  Only considers
// start tags and end tags with the name tagName when counting depth, so
// that the self-closing sort of tags don't interfere.
const char *skipThroughCloser(const char **posRef, const char *end, const char *tagName) {
	const char *pos = *posRef;
	const char *lastPosWithScoreOne = pos;
	int score = 1;
	for (;;) {
		const char *nextPos = eatSegment(pos, end);
		if (!nextPos) {
			*posRef = lastPosWithScoreOne;
			return NULL;
		}
		score += segmentTagDirection(pos, nextPos, tagName);
		if (score == 0) {
			*posRef = nextPos;
			return pos;
		}
		pos = nextPos;
		if (score == 1) {
			lastPosWithScoreOne = pos;
		}
	}
}

// stringPrefixWalk(beg, end, s)
//
// Returns NULL if s is not a prefix of [beg, end), returns beg + strlen(s)
// otherwise.
const char *stringPrefixWalk(const char *beg, const char *end, const char *s) {
	for (;;) {
		char c = *s;
		if (c == '\0') {
			return beg;
		}
		if (beg >= end || c != *beg) {
			return NULL;
		}
		++beg;
		++s;
	}
}

// regionEqualsString(beg, end, s)
//
// Returns YES iff [beg, end) is equal to [s, s + strlen(s)).
BOOL regionEqualsString(const char *beg, const char *end, const char *s) {
	return stringPrefixWalk(beg, end, s) == end;
}

// findString(beg, end, s)
//
// Inefficiently finds the substring s in the region [beg, end), returning
// the beginning of the substring, or NULL if the substring is not found.
const char *findString(const char *beg, const char *end, const char *s) {
	while (beg < end) {
		if (stringPrefixWalk(beg, end, s)) {
			return beg;
		}
		++beg;
	}
	return NULL;
}

// isTagWithAttr(beg, end, tagName, attrString, attrBegOut, attrEndOut)
//
// Returns YES if [beg, end) is a start tag with a substring exactly
// matching attrString.  attrString might equal 'foo="bar"' or it could
// just contain something more fragmentary than a name-value pair.
// Sets [*attrBegOut, *attrEndOut) to the region matching attrString.
BOOL isTagWithAttr(const char *beg, const char *end, const char *tagName, const char *attrString,
                   const char **attrBegOut, const char **attrEndOut) {
	if (!isOpeningTag(beg, end, tagName)) {
		return FALSE;
	}
	const char *attrBeg = findString(beg, end, attrString);
	if (!attrBeg) {
		return FALSE;
	}
	*attrBegOut = attrBeg;
	*attrEndOut = attrBeg + strlen(attrString);
	return TRUE;
}

// skipThroughTagWithAttr(posRef, end, tagName, attrString, attrBegOut, attrEndOut)
//
// Finds a tag with the given attribute, advancing *posRef in the process,
// returning success if it finds a tag with the matching attribute.  Sets
// [*attrBegOut, *attrEndOut) to the matching attribute and returns the
// beginning of the matching tag.  Or returns NULL if the tag has not yet
// been found.  Advances *posRef past the tag or as far as it can go.
const char *skipThroughTagWithAttr(const char **posRef, const char *end, const char *tagName, const char *attrString,
                                   const char **attrBegOut, const char **attrEndOut) {
	const char *pos = *posRef;
	for (;;) {
		const char *nextPos = eatSegment(pos, end);
		if (!nextPos) {
			*attrBegOut = NULL;
			*attrEndOut = NULL;
			*posRef = pos;
			return NULL;
		}
		if (isTagWithAttr(pos, nextPos, tagName, attrString, attrBegOut, attrEndOut)) {
			*posRef = nextPos;
			return pos;
		}
		pos = nextPos;
	}
}

// choiceTagWithAttr(posRef, end, tagName, attrString_A, attrString_B,
//                   attrBegOut, attrEndOut, A_wins)
//
// Like skipThroughTagWithAttr only with two possible attr strings.
// Sets [*attrBegOut, *attrEndOut) to the matching attribute string and
// sets A_wins to YES if attrString_A was matched.  Returns the beginning
// of the matching tag.  Returns NULL if no tag was matched.  Advances
// *posRef past the tag or as far as it can go.
const char *choiceTagWithAttr(const char **posRef, const char *end, const char *tagName,
                              const char *attrString_A, const char *attrString_B,
                              const char **attrBegOut, const char **attrEndOut,
                              BOOL *A_wins) {
	const char *pos = *posRef;
	for (;;) {
		const char *opener = skipThroughOpeningTag(&pos, end, tagName);
		if (!opener) {
			*posRef = pos;
			*attrBegOut = NULL;
			*attrEndOut = NULL;
			*A_wins = FALSE;  // it's false that A wins, it's false that B wins too..
			return NULL;
		}
		const char *beg_A = findString(opener, pos, attrString_A);
		if (beg_A) {
			*posRef = pos;
			*attrBegOut = beg_A;
			*attrEndOut = beg_A + strlen(attrString_A);
			*A_wins = TRUE;
			return opener;
		}
		const char *beg_B = findString(opener, pos, attrString_B);
		if (beg_B) {
			*posRef = pos;
			*attrBegOut = beg_B;
			*attrEndOut = beg_B + strlen(attrString_B);
			*A_wins = FALSE;
			return opener;
		}
	}
}

// skipThroughOpeningTag(posRef, end, tagName)
//
// Skips through an opening tag with the given name.  Advances *posRef past the
// tag or as far as it can go.  Returns the beginning of the matching tag or NULL.
const char *skipThroughOpeningTag(const char **posRef, const char *end, const char *tagName) {
	const char *pos = *posRef;
	for (;;) {
		const char *nextPos = eatSegment(pos, end);
		if (!nextPos) {
			*posRef = pos;
			return NULL;
		}
		if (isOpeningTag(pos, nextPos, tagName)) {
			*posRef = nextPos;
			return pos;
		}
		pos = nextPos;
	}
}

// isOpeningTag(beg, end, tagName)
//
// Returns YES if [beg, end) looks like a start tag with the given tag name.
BOOL isOpeningTag(const char *beg, const char *end, const char *tagName) {
	if (beg >= end || *beg != '<') {
		return FALSE;
	}
	const char *endOfTagName = stringPrefixWalk(beg + 1, end, tagName);
	return endOfTagName && endOfTagName < end && !isalpha(*endOfTagName);
}






